module Term: TERM =
  struct
	type term = Var of string | Cons of string * (term list) 
	type equationset = (term * term) list
	exception MatchFailure
	exception UndefinedMatch  
	exception UnifyFailure
	exception OccurCheck
	exception WrongSubstitution
	
	let rec ground = function
	  | Var _ -> false
	  | Cons(_, tl) -> groundl tl
	and groundl = function
	  | [] -> true
	  | t::tl1 -> ground(t) & groundl(tl1)

	  
	let rec occurs (stringa, termine) = match termine with
	   | Var st -> stringa = st
	   | Cons(_,tl) -> occursl (stringa, tl)
	and occursl (stringa, tl) = match tl with
	   | [] -> false
	   | t::tl1 -> occurs(stringa,t) or occursl(stringa, tl1)
	   
	let rec occursset (stringa, e) = match e with
	   | [] -> false
	   | (t1,t2)::e1 -> occurs(stringa, t1) or occurs(stringa, t2) 
	          or occursset(stringa, e1)
	
	let rec rplact (t, v, t1) = match t with
	  | Var v1 -> if v1 = v then t1 else t
	  | Cons(s,tl) -> Cons(s, rplactl(tl,v,t1))
	and rplactl (tl, v, t1) = match tl with
	  | [] -> []
	  | t::tl1 -> rplact(t, v, t1) :: rplactl(tl1, v, t1)
	
	let rec replace (eqset, stringa, termine) = match eqset with
	   | [] -> []
	   | (s1, t1) :: eqset1 -> 
	      (rplact(s1,stringa,termine),rplact(t1,stringa,termine))::
	       replace(eqset1, stringa, termine)
	       
	let rec pairs = function
	  | [], [] -> []
	  | a::l1, b::l2 -> (a,b)::pairs(l1, l2)
	  | _ -> raise UnifyFailure
	  
		  
	let rec unify  (eq1, eq2) = match eq1 with
  	   | [] -> eq2
  	   | (Var x, Var y) :: eq11 -> if x = y then unify(eq11, eq2) else
  	        unify(replace(eq11,x, Var y),(Var x, Var y)::(replace(eq2, x, Var y)))
  	   | (t, Var y) :: eq11 -> unify((Var y, t)::eq11,eq2)
  	   | (Var x, t) :: eq11 -> if occurs(x, t) then raise OccurCheck else
  	     unify(replace(eq11,x, t),(Var x, t)::(replace(eq2, x, t)))
  	   | (Cons(x,xl), Cons(y,yl)):: eq11 -> if not(x = y) then
  	        raise UnifyFailure else unify(pairs(xl,yl)@eq11, eq2)
	
	let unifyterms (t1, t2) = unify([(t1,t2)],[])
		
	let rec pmatch  (eq1, eq2) = match eq1 with
	   | [] -> eq2
  	   | (Var x, t) :: eq11 -> if occursset(x, eq11@eq2) then 
	     raise UndefinedMatch else pmatch(eq11,(Var x, t)::eq2)
  	   | (Cons(x,xl), Cons(y,yl)):: eq11 -> if not(x = y) then
  	        raise MatchFailure else pmatch(pairs(xl,yl)@eq11, eq2)
	   | _ -> raise UndefinedMatch
		
        let matchterms (pattern, termine) = 
	    if not (ground(termine))  then raise UndefinedMatch else
	    pmatch([(pattern,termine)],[])
	    
	let rec instantiate (t, e) = match e with
	  | [] -> t
	  | (Var x, t1) :: e1 -> instantiate (rplact(t, x, t1), e1) 
	  | _ -> raise WrongSubstitution
	      
	  
  end